# cartmodels.py -- Qt Model for manipulation of openbabel OBMol data
#
# Copyright (c) Adam Tenderholt, Stanford University, 2005-2007
#                               a-tenderholt@stanford.edu
# Copyright (c) Adam Tenderholt, atenderholt@gmail.com, 2012-2014
#
# This program is free software; you can redistribute it and/or modify  
# it under the terms of the GNU General Public License as published by 
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version with the exceptions noted in the 
# Help dialog.

from math import sin, cos, sqrt

from PyQt4 import QtCore, QtGui
import numpy

from cclib.parser import utils
table = utils.PeriodicTable()

class MoleculeModel(QtCore.QAbstractTableModel):
    def __init__(self, parent = None):
        QtCore.QAbstractTableModel.__init__(self,parent)

        self.atomcoords = None
        self.atomnos = None

    def headerData(self, section, orientation, role):
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            if section == 0:
                return QtCore.QVariant("Element")
            if section == 1:
                return QtCore.QVariant("x")
            if section == 2:
                return QtCore.QVariant("y")
            if section == 3:
                return QtCore.QVariant("z")

        if orientation == QtCore.Qt.Vertical and role == QtCore.Qt.DisplayRole:
            return QtCore.QVariant(section + 1)

        return QtCore.QVariant()

    def flags(self, index):
        if not index.isValid():
            return QtCore.Qt.ItemIsEnabled

        return QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEditable | \
                    QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsDragEnabled | \
                    QtCore.Qt.ItemIsDropEnabled

    def rowCount(self, parent = QtCore.QModelIndex()):
        if self.atomcoords != None:
            return self.atomcoords.shape[0]
        return 0

    def columnCount(self, parent = QtCore.QModelIndex()):
        if self.atomcoords != None:
            return self.atomcoords.shape[1] + 1
        return 0

    def data(self, index, role = QtCore.Qt.DisplayRole):
        if role != QtCore.Qt.DisplayRole and role != QtCore.Qt.EditRole:
            return QtCore.QVariant()

        row = index.row()
        col = index.column()

        if col < 0:
            print "Negative column!"
            return QtCore.QVariant()

        if col == 0:
            atomno = self.atomnos[row]
            name = table.element[atomno]
            return QtCore.QVariant(name)

        elif index.column() > 0:
            value = "%.3f" % self.atomcoords[row][col - 1]
            return QtCore.QVariant(value)

        return QtCore.QVariant("error")

    def setData(self, index, value, role):
        if role != QtCore.Qt.DisplayRole and role != QtCore.Qt.EditRole:
            return QtCore.QVariant()

        row = index.row()
        col = index.column()

        if col < 0:
            return False

        if col == 0:
            name = str(value.toString())
            num = table.number[name]
            self.atomnos[row] = num

        elif col > 0:
            pos = value.toDouble()[0]
            self.atomcoords[row][col - 1] = pos

        topLeft = self.index(0, 0)
        bottomRight = self.index(row, 3)

        self.emit(QtCore.SIGNAL("dataChanged(const QModelIndex&, const QModelIndex&)"), 
                    index, index)

        return True

    def centerMolecule(self, index):
        self.atomcoords -= self.atomcoords[index, :]

        topLeft = self.index(0,0)
        bottomRight = self.index(len(self.atomcoords),3)

        self.emit(QtCore.SIGNAL("dataChanged(const QModelIndex&, const QModelIndex&)"),
                    topLeft, bottomRight)

    def rotateMolecule(self, angle, rotvec, transvec = [0, 0, 0]):

        rotmat = numpy.zeros((3,3),"f")

        x = rotvec[0]
        y = rotvec[1]
        z = rotvec[2]

        length = sqrt(x**2 + y**2 + z**2)
        x /= length
        y /= length
        z /= length

        costh = cos(angle)
        sinth = sin(angle)

        rotmat[0,0] = costh + (1 - costh) * x**2
        rotmat[0,1] = (1 - costh) * x * y - sinth * z
        rotmat[0,2] = (1 - costh) * x * z + sinth*y
        rotmat[1,0] = (1 - costh) * x * y + sinth * z
        rotmat[1,1] = costh + (1 - costh) * y**2
        rotmat[1,2] = (1 - costh) * y * z - sinth * x
        rotmat[2,0] = (1 - costh) * z * x - sinth * y
        rotmat[2,1] = (1 - costh) * z * y + sinth * x
        rotmat[2,2] = costh + (1 - costh) * z**2

        temp = numpy.dot(rotmat, self.atomcoords.T)
        self.atomcoords = temp.T

        topLeft = self.index(0, 0)
        bottomRight = self.index(len(self.atomcoords), 3)

        self.emit(QtCore.SIGNAL("dataChanged(const QModelIndex&, const QModelIndex&)"),
                    topLeft, bottomRight)

        return self.atomcoords.copy()

    def slotTimer(self):
        self.rotateMolecule(10, [1, 0, 1])

    def setMolecule(self, atomcoords, atomnos):
        
        self.beginResetModel()
        self.atomcoords = atomcoords
        self.atomnos = atomnos
        self.endResetModel()

    def addAtom(self):
        size = self.obmol.NumAtoms()

        self.beginInsertRows(QtCore.QModelIndex(), size, size+1)
        newatom = self.obmol.NewAtom()
        newatom.SetAtomicNum(6)
        newatom.SetVector(0.0, 0.0, 0.0)
        self.endInsertRows()

        topLeft = self.index(size-1,0)
        bottomRight = self.index(size, 3)

        self.emit(QtCore.SIGNAL("dataChanged(const QModelIndex&, const QModelIndex&)"),topLeft, bottomRight)

    def removeAtom(self, i):

        self.beginRemoveRows(QtCore.QModelIndex(), i, i)
        atom = self.obmol.GetAtom(i+1)
        self.obmol.DeleteAtom(atom)
        self.endRemoveRows()

        size = self.obmol.NumAtoms()
        topLeft = self.index(i,0)
        bottomRight = self.index(size,3)
        self.emit(QtCore.SIGNAL("dataChanged(const QModelIndex&, const QModelIndex&)"),topLeft, bottomRight)

if __name__ == '__main__':
    import sys
    from cclib.parser import ccopen
    
    app = QtGui.QApplication(sys.argv)

    parser = ccopen(sys.argv[1])
    data = parser.parse()

    view = QtGui.QTableView()
    view.setDragEnabled(True)
    view.setAcceptDrops(True)
    view.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)

    model = MoleculeModel()
    model.atomcoords = data.atomcoords[0]
    model.atomnos = data.atomnos

    view.setModel(model)
    view.setWindowTitle("Atoms")
    view.show()

#    timer = QtCore.QTimer()
#    QtCore.QObject.connect(timer, QtCore.SIGNAL("timeout()"),model.slotTimer)
#    timer.start(1000)
#
#    glwidget = GLWidget()
#    glwidget.obmol = molecule
#    glwidget.show()
#
#    QtCore.QObject.connect(model,QtCore.SIGNAL("dataChanged(const QModelIndex&, const QModelIndex&)"),glwidget.slotDataChanged)

    sys.exit(app.exec_())
