
# cartwidget.py -- widget used in the cartesion coord editor tab
#
# Copyright (c) Adam Tenderholt, Stanford University, 2005-2006
#                               a-tenderholt@stanford.edu
# Copyright (c) Adam Tenderholt, atenderholt@gmail.com, 2012-2014
#
# This program is free software; you can redistribute it and/or modify  
# it under the terms of the GNU General Public License as published by 
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version with the exceptions noted in the 
# Help dialog.

from math import sin, cos, sqrt, acos
import numpy

from PyQt4 import QtGui,QtCore
from cclib.parser import utils

from glwidget import GLWidget
from customdelegate import CustomDelegate
from cartmodels import MoleculeModel
from dialogs import SaveFormatDialog
from ui_cartesian_widget import Ui_cartWidget

table = utils.PeriodicTable()

class CartWidget(QtGui.QWidget, Ui_cartWidget):
    
    def __init__(self, parent = None, flags = 0):
        super(CartWidget, self).__init__(parent)
        self.setupUi(self)

        self.cartModel = MoleculeModel()
        self.cartView.setModel(self.cartModel)
        self.cartView.show()

        self.customDelegate = CustomDelegate()
        self.selectionModel = QtGui.QItemSelectionModel(self.cartModel)
        self.cartView.setSelectionModel(self.selectionModel)

#slots
        selection_changed_str = "selectionChanged(const QItemSelection&, const QItemSelection&)"
        QtCore.QObject.connect(self.selectionModel, QtCore.SIGNAL(selection_changed_str), 
                                self.changeCartSelection)
        QtCore.QObject.connect(self.glWidget,QtCore.SIGNAL("signalRotateMolecule"), 
                                self.cartModel.rotateMolecule)
        data_changed_str = "dataChanged(const QModelIndex&, const QModelIndex&)"
        QtCore.QObject.connect(self.cartModel, QtCore.SIGNAL(data_changed_str),
                                self.slotUpdateCartData)

#class variables
        self.data = None

    def setData(self,data):
        self.data = data

        if not hasattr(data, "atomcoords"):
            return

        if not hasattr(data, "atomnos"):
            return

        size = len(data.atomcoords)
        self.spinBox.setMaximum(size - 1)
        
        self.cartModel.setMolecule(data.atomcoords[0].copy(), data.atomnos.copy())

        for i in range(size):
            self.glWidget.appendMolecule(data.atomcoords[i].copy(), data.atomnos.copy())

        self.cartView.resizeColumnsToContents()

    def setMolecule(self, molecule):

        self.cartModel.setMolecule(molecule)
        self.cartView.setModel(self.cartModel)
        self.glWidget.obmol = self.cartModel.obmol

        self.cartView.setColumnWidth(0,55)
        self.cartView.setColumnWidth(1,65)
        self.cartView.setColumnWidth(2,65)
        self.cartView.setColumnWidth(3,65)
        self.cartView.setItemDelegate(self.customDelegate)

    def slotUpdateGeometry(self, value):

        if not self.data:
            return

        self.cartModel.setMolecule(self.data.atomcoords[value], self.data.atomnos)
        self.glWidget.current = value

        self.glWidget.updateGL()
        self.spinBox.setValue(value)

        self.emit(QtCore.SIGNAL("signalUpdateGeometry(int)"),value)

    def changeCartSelection(self, current, previous):
        indexes = self.selectionModel.selectedIndexes()
        dict = {}
        for index in indexes:
            dict[index.row()] = 1
        self.glWidget.setSelected(dict.keys())

    def slotUpdateCartData(self, topLeft, bottomRight):

        row = topLeft.row()
        col = topLeft.column()

        data = self.cartModel.data(topLeft)

        if col == 0:
            tempdata = data.toString()
            data = table.number[str(tempdata)]

        else:
            data, ok = data.toDouble()

        self.glWidget.updateMolecule(row, col, data)

    def centerMolecule(self):
        indexes = self.cartView.selectedIndexes()

        if len(indexes) != 1:
            return

        index = indexes[0].row()
        self.cartModel.centerMolecule(index)
        self.glWidget.centerMolecule(index)

    def rotateToX(self):
        mask = [0.0, 1.0, -1.0]
        indices = [0, 2, 1]
        self.rotateToVector(mask, indices, 0)

    def rotateToY(self):
        mask = [-1.0, 0.0, 1.0]
        indices = [2, 1, 0]
        self.rotateToVector(mask, indices, 1)

    def rotateToZ(self):
        mask = [1.0, -1.0, 0]
        indices = [1, 0, 2]
        self.rotateToVector(mask, indices, 2)

    def rotateToVector(self, mask, indices, angle_index):
        indexes = self.cartView.selectedIndexes()
        
        if len(indexes) != 2:
            return

        index1 = indexes[0].row()
        index2 = indexes[1].row()
        if index1 == index2:
            return

        atomcoords = self.cartModel.atomcoords
        vec = atomcoords[index2, :]  - atomcoords[index1, :]

        length = sqrt(numpy.dot(vec, vec))
        rotvec = vec[indices] * mask / length

        angle = acos(vec[angle_index]/length)
        atomcoords = self.cartModel.rotateMolecule(angle, [rotvec[0], rotvec[1], rotvec[2]])
        atomnos = self.cartModel.atomnos
        current = self.spinBox.value()

        self.glWidget.updateBondCallList(current, atomcoords, atomnos)
        self.glWidget.atomcoords[current] = atomcoords
        self.glWidget.update()

    def rotateToXY(self):
        mask = [1.0, -1.0, 0.0]
        indices = [1, 0, 2]
        self.rotateToPlane(mask, indices, 2)

    def rotateToXZ(self):
        mask = [-1.0, 0.0, 1.0]
        indices = [2, 1, 0]
        self.rotateToPlane(mask, indices, 1)

    def rotateToYZ(self):
        mask = [0.0, 1.0, -1.0]
        indices = [0, 2, 1]
        self.rotateToPlane(mask, indices, 0)

    def rotateToPlane(self, mask, indices, angle_index):
        indexes = self.cartView.selectedIndexes()
        
        if len(indexes) != 3:
            return

        index1 = indexes[0].row()
        index2 = indexes[1].row()
        index3 = indexes[2].row()

        if index1 == index2:
            return
        if index2 == index3:
            return
        if index1 == index3:
            return

        atomcoords = self.cartModel.atomcoords
        vec1 = atomcoords[index2, :]  - atomcoords[index1, :]
        vec2 = atomcoords[index3, :]  - atomcoords[index1, :]
        vec = numpy.cross(vec1, vec2)

        length = sqrt(vec[0]**2 + vec[1]**2 + vec[2]**2)
        rotvec = vec[indices] * mask / length
        angle = acos(vec[angle_index]/length)

        atomcoords = self.cartModel.rotateMolecule(angle, [rotvec[0], rotvec[1], rotvec[2]])
        atomnos = self.cartModel.atomnos
        current = self.spinBox.value()

        self.glWidget.updateBondCallList(current, atomcoords, atomnos)
        self.glWidget.atomcoords[current] = atomcoords
        self.glWidget.update()

    def addAtom(self):
        self.cartModel.addAtom()
        self.glWidget.obmol = self.cartModel.obmol
        self.cartView.setColumnWidth(0,55)
        self.cartView.setColumnWidth(1,65)
        self.cartView.setColumnWidth(2,65)
        self.cartView.setColumnWidth(3,65)
        self.cartView.setItemDelegate(self.customDelegate)

    def removeAtom(self):
        indexes = self.selectionModel.selectedIndexes()

        #get indices and sort
        rows = []
        for index in indexes:
            rows.append(index.row())
        rows.sort()

        i = 0
        for row in rows:
            self.cartModel.removeAtom(row - i)
            i += 1

if __name__=='__main__':
    import sys
    from os.path import exists
    from cclib.parser import ccopen

    app = QtGui.QApplication(sys.argv)
    window = CartWidget()
    window.resize(700,500)
    window.show()
    button = QtGui.QPushButton("foo")
    button.show()
    QtCore.QObject.connect(button, QtCore.SIGNAL("clicked()"), window.addAtom)

    if len(sys.argv)==2 and exists(sys.argv[1]):
        parser = ccopen(sys.argv[1])
        data = parser.parse()
        window.setData(data) 

    sys.exit(app.exec_())

