
# formatter.py -- code for formatting the results of various analyses
#
# Copyright (c) Adam Tenderholt, Stanford University, 2005-2007
#                               a-tenderholt@stanford.edu
# Copyright (c) Adam Tenderholt, atenderholt@gmail.com, 2012-2014
#
# This program is free software; you can redistribute it and/or modify  
# it under the terms of the GNU General Public License as published by 
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version with the exceptions noted in the 
# Help dialog.

from time import asctime
from math import fabs

import numpy

from cclib.parser import utils

class Formatter(object):
    def __init__(self, filename):
        self.filename = filename

    def printCDA(self, data, names, paths, method):

        allText = "Charge decomposition analysis of %s "%self.filename
        allText += "using the following fragments:\n"

        for i in range(len(names)):
            allText += "%s (%s)\n"%(names[i], paths[i])

        allText += "\nCreated on "+asctime()+"\n\n"

        if len(data.homos) == 2:
            allText += "ALPHA SPIN:\n"
            allText += "===========\n"

        allText += " MO#      d       b       r\n"
        allText += "-----------------------------\n"

        for spin in range(len(data.homos)):

            if spin == 1:
                allText += "\nBETA SPIN:\n"
                allText += "==========\n"
                allText += " MO#      d       b       r\n"
                allText += "-----------------------------\n"

            for i in range(len(method.donations[spin])):

                allText += "%4i: %7.3f %7.3f %7.3f\n"%(i+1,2*method.donations[spin][i], \
                                                    2*method.bdonations[spin][i], \
                                                    2*method.repulsions[spin][i])

                if i == data.homos[spin]:
                    allText += "------ HOMO - LUMO gap ------\n"

                if i > data.homos[spin] + 2:
                    allText += ".......\n"
                    break

            allText += "%4i: %7.3f %7.3f %7.3f\n"%(len(method.donations[spin]),
                                                2*method.donations[spin][-1], \
                                                2*method.bdonations[spin][-1], \
                                                2*method.repulsions[spin][-1])


            allText += "-----------------------------\n"
            allText += " T:   %7.3f %7.3f %7.3f\n"%(reduce(numpy.add,\
                                            method.donations[spin])*2, \
                                            reduce(numpy.add, method.bdonations[spin])*2, \
                                            reduce(numpy.add, method.repulsions[spin])*2)

        return allText
        
    def printExtendedPA(self, data, fragmethod, method, type):
        allText = ""
        allText += self.printHeader(type)

        if len(data.moenergies) == 2:
            allText+="ALPHA SPIN:\n"

        for spin in range(len(data.moenergies)):

            if spin == 1:
                allText+="\nBETA SPIN:\n"

            allText += "%3s %12s    Composition\n"%("MO","Energy(eV)")
            allText += "-"*72 + "\n"
               
            for i in range(len(data.moenergies[spin])):
                text = ""
                sorted = numpy.sort(method.aoresults[spin][i])
                indexes = numpy.argsort(method.aoresults[spin][i])
                
                if i == data.homos[spin] + 1:
                    allText += "--- HOMO - LUMO gap --- %9.3f eV ----"%(data.moenergies[spin][i]-data.moenergies[spin][i-1],)
                    allText += "-"*32 + "\n"
                text += "%3i %12.3f    "%(i+1, data.moenergies[spin][i])
                size = len(method.aoresults[spin][i])
                for j in range(size):
                    val = sorted[size-1-j]
                    index = indexes[size-1-j]
                    if fabs(val) > 0.01:
                        if len(text) > 70:
                            allText += text + "\n"
                            text = "                    "
                        text += "%.2f%% %s, "%(val*100, fragmethod.fonames[index])

                text = text[:-2] #remove last comma
                allText += text + "\n"

        return allText

    def printAllPA(self, data, fraglists, fragnames, method, type):
        allText = ""
        allText+=self.printHeader(type)
        allText+=self.printFragList(fraglists, fragnames)

        if len(data.moenergies) == 2:
            allText+="ALPHA SPIN:\n"

        allText+=self.printPAKey(method.fragresults[0])
        if not hasattr(data,"mosyms"):
            syms = ['A'] * len(data.moenergies[0])
            allText += self.printResults(data.moenergies[0], syms, data.homos[0],\
                                            method.fragresults[0])
        else:
            allText+=self.printResults(data.moenergies[0], data.mosyms[0], \
                                        data.homos[0],method.fragresults[0])

        if len(data.moenergies) == 2:
            allText+="\nBETA SPIN:\n"

            allText+=self.printPAKey(method.fragresults[1])

            if not hasattr(data,"mosyms"):
                allText += self.printResults(data.moenergies[1], syms, data.homos[1], \
                                                method.fragresults[1])
            else:
                allText+=self.printResults(data.moenergies[1], data.mosyms[1], \
                                                data.homos[1], method.fragresults[1])

        return allText

    def printAllOP(self,data,fraglists,fragnames,method,type):
        
        allText=self.printHeader(type)
        allText+=self.printFragList(fraglists,fragnames)
        
        
        for spin in range(len(method.results)):

            if len(data.moenergies) == 2 and spin == 0:
                allText+="ALPHA SPIN:\n"

            elif spin == 1:
                allText+="\n\nBETA SPIN:\n"
                
#print the alpha/restricted results
            allText+=self.printOPKey(method.results[spin])
            if not hasattr(data,"mosyms"):
                symms = ['A'] * len(data.moenergies[spin])
            else:
                symms = data.mosyms[spin]

            mainString=""
            homo = data.homos[spin]
            lumo=homo+1

            for i in range(len(method.results[spin])):
                
                if(homo-i<20 and i<lumo):
                    string="%4i %3i %9.3f %5s "%(i+1,i-homo,data.moenergies[spin][i],symms[i])
                elif(i-lumo<20 and i>homo):
                    string="%4i %3i %9.3f %5s "%(i+1,i-lumo,data.moenergies[spin][i],symms[i])
                else:
                    string="%4i     %9.3f %5s "%(i+1,data.moenergies[spin][i],symms[i])
                
                if(i==lumo):
                    mainString+="--- HOMO - LUMO gap --- %9.3f eV ----"%(data.moenergies[spin][i]-data.moenergies[spin][i-1],)

                    if(ncols>2):
                        mainString+=(ncols-2)*"--------"
                            
                    mainString+='\n'
                    
                ncols=0 #used for keeping track of number of cols
                #loop through each combinatin of fragments
                for a in range(len(fragnames)-1): 
                    for b in range(a+1,len(fragnames)):
                        string+="%7.3f "% method.results[spin][i][a][b]
                        ncols += 1
                    
                mainString+=string+"\n"

            allText+=mainString

        return allText

#print the beta results
        if len(data.moenergies) == 2:
            allText+="\nBETA SPIN:\n"

            allText+=self.printPAKey(method.results[1])

            if not hasattr(data,"mosyms"):
                allText += self.printResults(data.moenergies[1], syms, data.homos[1], \
                                                method.results[1])
            else:
                allText+=self.printResults(data.moenergies[1], data.mosyms[1], \
                                                data.homos[1], method.results[1])
       
        return allText

    def printGrossPopulation(self, data, mpa, lpa):
        allText = self.printHeader("Gross Population Analysis")

        # get list of atoms
        t = utils.PeriodicTable()
        atoms = [ t.element[i] for i in data.atomnos ]

        allText += self.printSubFragPopulation(data, atoms,
                                                mpa.fragresults, lpa.fragresults,
                                                "atom")
        allText += "\n\n"
        allText += self.printSubFragPopulation(data, data.aonames,
                                                mpa.aoresults, lpa.aoresults,
                                                "orbital")
        return allText

    def printSubFragPopulation(self, data, frags, mpa, lpa, type):
        # determine whether spin-restricted or not
        alpha = 0
        if len(data.homos) == 1:
            beta = 0
        else:
            beta = 1

        lumo0 = data.homos[alpha] + 1
        lumo1 = data.homos[beta] + 1

        # first calculate mulliken atom populations
        mpa0 = mpa[alpha][0:lumo0, :].sum(axis=0)
        mpa1 = mpa[beta][0:lumo1, :].sum(axis=0)
        mpa_total = mpa0 + mpa1
        mpa_spin = mpa0 - mpa1

        # and now calculate lowdin atom populations
        lpa0 = lpa[alpha][0:lumo0, :].sum(axis=0)
        lpa1 = lpa[beta][0:lumo1, :].sum(axis=0)
        lpa_total = lpa0 + lpa1
        lpa_spin = lpa0 - lpa1

        if type == "atom":
            allText = "       "
            allText += "  --------- Mulliken ---------"
            allText += "  ---------- Lowdin ----------\n"
            allText += " %6s  %5s  %5s  %5s  %6s   %5s  %5s  %5s  %6s\n" % \
                            ("# Atom", "Alpha", "Beta", "Total", "Spin", "Alpha", "Beta",
                                "Total","Spin")
        elif type == "orbital":
            allText = "                "
            allText += "  --------- Mulliken ---------"
            allText += "  ---------- Lowdin ----------\n"
            allText += "%4s %-10s   %5s  %5s  %5s  %6s   %5s  %5s  %5s  %6s\n" % \
                            ("#", "Orbital", "Alpha", "Beta", "Total", "Spin", "Alpha", "Beta",
                                "Total","Spin")


        for i, frag in enumerate(frags):
            if type == "atom":
                subText = "%3i %-2s:" % (i + 1, frag)
            elif type == "orbital":
                subText = "%4i %-10s:" % (i + 1, frag)

            subText += "  %5.2f  %5.2f  %5.2f  %7.4f" % (mpa0[i], mpa1[i], mpa_total[i],
                                                            mpa_spin[i])
            subText += "  %5.2f  %5.2f  %5.2f  %7.4f\n" % (lpa0[i], lpa1[i], lpa_total[i],
                                                            lpa_spin[i])
            allText += subText

        return allText

    def printHeader(self,type):
    
        headerText="Results from "+type+" of "+self.filename+"\n"
        headerText+="Created on "+asctime()+"\n\n"
        
        return headerText

    def printFragList(self,fraglists,fragnames):
        
        fragText=""
        
        for i in range(len(fraglists)):
            
            string=self.getStringRep(fraglists[i])
            fragText+="Frag. %i (%s): Orbitals %s\n"%(i+1, fragnames[i],string)
            
        fragText+="\n"
        
        return fragText
        
    def printPAKey(self,results):
        
        string="%4s %3s %9s %5s    "%("MO","","Energy(eV)","Sym.")
        string2="--------------------------"
        for i in range(len(results[-1])):
            string+="%-6s  "% (i+1,)
            string2+="--------"
            
        return string+"\n"+string2+"\n"

    def printOPKey(self,results):
        
        string="%4s %3s %9s %5s "%("MO","","Energy(eV)","Sym.")
        string2="--------------------------"
        for i in range(len(results[-1])-1):
            
            for j in range(i+1,len(results[-1])):
                string+=" %2s&%-2s  "%(i+1, j+1)
                string2+="--------"
            
        return string+"\n"+string2+"\n"

    def printResults(self,eigens,symms,homo,results,scale=100):
    
        mainString=""
        lumo=homo+1
        
        if len(results) > len(eigens):
            size = len(eigens)
        else:
            size = len(results)

        #print formatted results
        for i in range(size):

            if eigens[i] == 99999:
                continue
            
            if(homo-i<20 and i<lumo):
                string="%4i %3i %9.3f %5s "%(i+1,i-homo,eigens[i],symms[i])
            elif(i-lumo<20 and i>homo):
                string="%4i %3i %9.3f %5s "%(i+1,i-lumo,eigens[i],symms[i])
            else:
                string="%4i     %9.3f %5s "%(i+1,eigens[i],symms[i])
            
            if(i==lumo):
                mainString+="--- HOMO - LUMO gap --- %9.3f eV --------"%(eigens[i]-eigens[i-1],)
                size=len(results[-1])
                if(size>3):
                    if(scale==100):
                        mainString+=(size-3)*"-------"
                    else:
                        mainString+=(size-3)*"--------"+"---"
                        
                mainString+='\n'
                
            for result in results[i]:
                if scale==100:
                    string+="%7.2f "%(result*scale,)
                else:
                    string+="%7.3f "%(result*scale,)
                
            mainString+=string+"\n"
           
        return mainString

#take a list of numbers, and convert to a convenient string        
    def getStringRep(self,list):
        cont=[]
        temp=[]
        list.sort()
        temp.append(list[0])
        cont.append(temp)
        list.pop(0)
        
        #go through list and move continuous numbers into sublists
        while(len(list)>0):
            if(list[0]-cont[-1][-1]==1):
                temp.append(list[0])
            else:
                temp=[]
                cont.append(temp)
                temp.append(list[0])
            list.pop(0)
        
        #look at sublists, and based on size, determine how to format
        string=""
        for list in cont[:-1]:
            if(len(list)==1):
                string+=str(list[0]+1)+", "
            else:
                string+=str(list[0]+1)+"-"+str(list[-1]+1)+", "
                
        if(len(cont[-1])==1):
            string+=str(cont[-1][0]+1)
        else:
            string+=str(cont[-1][0]+1)+"-"+str(cont[-1][-1]+1)
                
        return string
  
if __name__=='__main__':
    app = QtGui.QApplication(sys.argv)
    window = ResultsDialog()
    window.show()

    sys.exit(app.exec_())

