# geomwidget.py -- widget used in the geometry optimization tab
#
# Copyright (c) Adam Tenderholt, Stanford University, 2005-2006
#                               a-tenderholt@stanford.edu
# Copyright (c) Adam Tenderholt, atenderholt@gmail.com, 2012-2014
#
# This program is free software; you can redistribute it and/or modify  
# it under the terms of the GNU General Public License as published by 
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version with the exceptions noted in the 
# Help dialog.

from PyQt4 import QtGui, QtCore, Qwt5
import numpy

from glwidget import GLWidget
from dialogs import SaveFormatDialog
from ui_geometry_widget import Ui_geomWidget

class GeomWidget(QtGui.QWidget, Ui_geomWidget):
    
    def __init__(self, parent = None, flags = 0):
        super(GeomWidget, self).__init__(parent)
        self.setupUi(self)

        # setup energy plot labels/curves
        self.energyPlot.setAxisTitle(Qwt5.QwtPlot.xBottom, "Step")
        self.energyPlot.setAxisTitle(Qwt5.QwtPlot.yLeft, "Energy (eV)")
        self.energyCurve = Qwt5.QwtPlotCurve()
        self.energyCurve.attach(self.energyPlot)
        self.energySelected = Qwt5.QwtPlotCurve()
        self.energySelected.attach(self.energyPlot)

        # setup convergence plot labels/curves
        self.convergencePlot.setAxisTitle(Qwt5.QwtPlot.xBottom, "Step")
        self.convergencePlot.setAxisTitle(Qwt5.QwtPlot.yLeft, "Convergence")
        self.convergenceCurve = Qwt5.QwtPlotCurve()
        self.convergenceCurve.attach(self.convergencePlot)
        self.convergenceSelected = Qwt5.QwtPlotCurve()
        self.convergenceSelected.attach(self.convergencePlot)

        # setup brushes
        self.symbol = Qwt5.QwtSymbol()
        self.symbol.setStyle(Qwt5.QwtSymbol.XCross)
        self.symbol.setSize(7)
        self.selectedSymbol = Qwt5.QwtSymbol(self.symbol)
        pen = QtGui.QPen(QtGui.QColor("red"))
        pen.setWidth(2)
        self.selectedSymbol.setPen(pen)

        # assign symbols
        self.energyCurve.setSymbol(self.symbol)
        self.energySelected.setSymbol(self.selectedSymbol)
        self.convergenceCurve.setSymbol(self.symbol)
        self.convergenceSelected.setSymbol(self.selectedSymbol)

        # setup pickers
        self.energyPicker = Qwt5.QwtPlotPicker(self.energyPlot.canvas())
        self.energyPicker.setRubberBand(Qwt5.QwtPlotPicker.CrossRubberBand)
        self.energyPicker.setTrackerMode(Qwt5.QwtPicker.AlwaysOn)
        self.energyPicker.setSelectionFlags(Qwt5.QwtPicker.PointSelection | \
                                                Qwt5.QwtPicker.DragSelection)
        self.convergencePicker = Qwt5.QwtPlotPicker(self.convergencePlot.canvas())
        self.convergencePicker.setRubberBand(Qwt5.QwtPlotPicker.CrossRubberBand)
        self.convergencePicker.setTrackerMode(Qwt5.QwtPicker.AlwaysOn)
        self.convergencePicker.setSelectionFlags(Qwt5.QwtPicker.PointSelection | \
                                                Qwt5.QwtPicker.DragSelection)


        QtCore.QObject.connect(self.energyPicker, QtCore.SIGNAL("moved(const QwtDoublePoint&)"), 
                                self.slotSelected)
        QtCore.QObject.connect(self.convergencePicker, 
                                QtCore.SIGNAL("moved(const QwtDoublePoint&)"), 
                                self.slotSelected)

        self.data = None

    def slotSelected(self, point):
        if not hasattr(self.data, "atomcoords"):
            return

        rounded_point = round(point.x())
        if rounded_point >= 0 and rounded_point < len(self.data.atomcoords):
            self.slotUpdateGeometry(rounded_point)

    def setData(self,data):
        self.data = data

        if not hasattr(data, "atomcoords"):
            return

        if not hasattr(data, "atomnos"):
            return

        size = len(data.atomcoords)
        self.spinBox.setMaximum(size - 1)
        self.slider.setMaximum(size - 1)
        
        for i in range(len(data.atomcoords)):
            self.glWidget.appendMolecule(data.atomcoords[i], data.atomnos)

        # plot energies if available
        if hasattr(data, "scfenergies"):
            xvalues = range(len(data.scfenergies))
            yvalues = (data.scfenergies - data.scfenergies[0])
            self.energyCurve.setData(xvalues, yvalues)
            self.energySelected.setData(xvalues[:1], yvalues[:1])
            self.energyPlot.replot()

        if hasattr(data, "geovalues") and hasattr(data, "geotargets"):
            self.deviation = []
            for step in data.geovalues:
                filter = step > data.geotargets
                self.deviation.append(0 + numpy.log(step[filter] / data.geotargets[filter]).sum())
            if len(self.deviation) < len(data.scfenergies):
                self.deviation.append(0.0)

            xvalues = range(len(self.deviation))
            self.convergenceCurve.setData(xvalues, self.deviation)
            self.convergenceSelected.setData(xvalues[:1], self.deviation[:1])
            self.convergencePlot.replot()

    def slotUpdateGeometry(self, value):
        value = int(value)

        if not self.data:
            return

        self.slider.setValue(value)
        self.spinBox.setValue(value)
        self.glWidget.current = value
        self.glWidget.update()

        if hasattr(self.data, "scfenergies"):
            energy = self.data.scfenergies[value] - self.data.scfenergies[0]
            self.energySelected.setData([value], [energy])
            self.energyPlot.replot()

        if hasattr(self, "deviation"):
            self.convergenceSelected.setData([value], [self.deviation[value]])
            self.convergencePlot.replot()

        self.emit(QtCore.SIGNAL("signalUpdateGeometry(int)"),value)

if __name__=='__main__':
    import sys
    from os.path import exists

    from cclib.parser import ccopen

    app = QtGui.QApplication(sys.argv)
    window = GeomWidget()
    window.resize(700,500)
    window.show()
    if len(sys.argv)==2 and exists(sys.argv[1]):
        parser = ccopen(sys.argv[1])
        data = parser.parse()
        window.setData(data) 

    sys.exit(app.exec_())

