# glwidget.py -- widget used to display molecules in 3d
#
# Copyright (c) Adam Tenderholt, Stanford University, 2005-2006
#                               a-tenderholt@stanford.edu
# Copyright (c) Adam Tenderholt, atenderholt@gmail.com, 2012-2014
#
# This program is free software; you can redistribute it and/or modify  
# it under the terms of the GNU General Public License as published by 
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version with the exceptions noted in the 
# Help dialog.

from math import sqrt,cos,sin,tan,acos,pi
from ctypes import util

from PyQt4 import QtOpenGL
from PyQt4 import QtGui,QtCore
from OpenGL.GL import *
from OpenGL.GLU import *

from elements import get_elements
etable = get_elements()

LENGTH, ANGLE, TORSION = range(3)
BOND_TOLERANCE = 0.45

class GLWidget(QtOpenGL.QGLWidget):
    def __init__(self, parent = None):
        QtOpenGL.QGLWidget.__init__(self,parent)

        self.bond_call_lists = {}
        self.atom_call_lists = {}
        self.xrot = 0
        self.yrot = 0
        self.zrot = 0
        self.xpos = 0.0
        self.ypos = 0.0
        self.zpos = -15.0
        self.current = 0
        self.lastPos = QtCore.QPoint()
        self.locked = True
        self.selected = []

        self.atomcoords = {}
        self.atomnos = {}

    def slotDataChanged(self, topLeft, bottomRight):
        self.updateGL()

    def setSelected(self,list):
        self.selected = list
        self.updateGL()

    def slotEnableLock(self,value):
        self.locked = ( value == QtCore.Qt.Checked )

    def wheelEvent(self, event):
        if event.orientation() == QtCore.Qt.Horizontal:
            pass
        else:
            self.zpos += event.delta() / 20.0
            if self.zpos > -5.0:
                self.zpos = -5.0
            if self.zpos < -100.0:
                self.zpos = -100.0

        event.accept()
        self.updateGL()

    def mousePressEvent(self, event):
        self.lastPos = QtCore.QPoint(event.pos())

    def mouseMoveEvent(self, event):
        dx = event.x() - self.lastPos.x()
        dy = event.y() - self.lastPos.y()
        
        if event.buttons() == QtCore.Qt.LeftButton:
            if self.locked:
                self.setXRotation(self.xrot + 8 * dy)
                self.setYRotation(self.yrot + 8 * dx)
            else:
                #determine perp vector
                length = sqrt(dx**2 + dy**2)
                x = -dy/length
                y = dx/length

                angle = length/180 * pi
                costh = cos(angle)
                sinth = sin(angle)

                self.emit(QtCore.SIGNAL("signalRotateMolecule"),angle, [x, y, 0.0])

        elif event.buttons() == QtCore.Qt.RightButton:
            self.xpos += dx / 10.0
            self.ypos -= dy / 10.0
            
        self.lastPos = QtCore.QPoint(event.pos())
        self.updateGL()

    def center(self):
        self.xpos = 0.0
        self.ypox = 0.0
        self.zpos = -15.0

        self.updateGL()

    def orient(self, x, y, z):
        self.xrot = x * 16
        self.yrot = y * 16
        self.zrot = z * 16
        
        self.updateGL()

    def setXRotation(self, angle):
        angle = self.normalizeAngle(angle)
        if angle != self.xrot:
            self.xrot = angle
            self.emit(QtCore.SIGNAL("xRotationChanged(int)"),angle)

    def setYRotation(self, angle):
        angle = self.normalizeAngle(angle)
        if angle != self.yrot:
            self.yrot = angle
            self.emit(QtCore.SIGNAL("yRotationChanged(int)"),angle)

    def setZRotation(self, angle):
        angle = self.normalizeAngle(angle)
        if angle != self.zrot:
            self.zrot = angle
            self.emit(QtCore.SIGNAL("zRotationChanged(int)"),angle)

    def normalizeAngle(self, angle):
        while angle < 0:
            angle += 360 * 16
        while angle > 360 * 16:
            angle -= 360 * 16
        return angle

    def removeMolecule(self, key):
        if key not in self.bond_call_lists:
            return

        call_list = self.bond_call_lists[key]
        glDeleteLists(call_list, 1)

        del self.atomcoords[key]
        del self.atomnos[key]

    def reset(self):
        for key in self.bond_call_lists.keys():
            self.removeMolecule(key)

        self.bond_call_lists.clear()

    def updateAtomCallLists(self, elements):
        self.makeCurrent()

        # setup the quadrics first
        quadric = gluNewQuadric()
        gluQuadricNormals(quadric, GLU_SMOOTH)

        # now go through and update call lists if atoms are absent
        elements = map(int, elements)
        for i, element in enumerate(elements):

            # skip if element type has already been set in call list
            if element in self.atom_call_lists:
                continue

            radius = etable[element]['vradius']
            rgb = etable[element]['rgb']

            # create call lists
            start_list = glGenLists(2)
            self.atom_call_lists[element] = (start_list, start_list + 1)

            glNewList(start_list, GL_COMPILE)

            glColor4f(rgb[0], rgb[1], rgb[2], 1.0)
            gluSphere(quadric, radius*0.2, 16, 16)
            
            glEndList()

            # and now create call list for selected atoms
            glNewList(start_list + 1, GL_COMPILE)

            glColor4f(1.0, 1.0, 1.0, 0.5)
            gluSphere(quadric, radius*0.3, 16, 16)

            glEndList()

        gluDeleteQuadric(quadric)

    def updateBondCallList(self, index, atomcoords, atomnos):
        self.makeCurrent()

        # setup the quadrics first
        quadric = gluNewQuadric()
        gluQuadricNormals(quadric, GLU_SMOOTH)

        glNewList(self.bond_call_lists[index], GL_COMPILE)

        for i, coord1 in enumerate(atomcoords[:-1]):
            num1 = atomnos[i]
            color1 = etable[num1]['rgb']
            radius1 = etable[num1]['cradius']

            for j, coord2 in enumerate(atomcoords[i + 1:], i + 1):
                num2 = atomnos[j]
                color2 = etable[num2]['rgb']
                radius2 = etable[num2]['cradius']

                deltas = coord2 - coord1
                length = sqrt(sum( [ x * x for x in deltas ] ))

                if length >= radius1 + radius2 + BOND_TOLERANCE:
                    continue

                # apparently some corrections
                prerot = [ -1*deltas[1], deltas[0], 0.0 ]
                rotsize = sqrt(sum( [ x * x for x in prerot ] ))
                if rotsize != 0:
                    rot = [prerot[0] / rotsize, prerot[1] / rotsize, 0.0] 
                    angle = acos(deltas[2]/length)/pi*180
                elif deltas[2] < 0:
                    rot = [1.0, 0.0, 0.0]
                    angle = 180
                else:
                    rot = [1.0, 0.0, 0.0]
                    angle = 0

                # actual rendering
                glPushMatrix()

                # first half
                glTranslatef(coord1[0], coord1[1], coord1[2])
                glRotatef(angle, rot[0], rot[1], rot[2])
                glColor4f(color1[0], color1[1], color1[2],1.0)
                gluCylinder(quadric, 0.15, 0.15, length/2, 16, 8)

                # second half
                glTranslatef(0, 0, length / 2)
                glColor4f(color2[0], color2[1], color2[2], 1.0)
                gluCylinder(quadric, 0.15, 0.15, length/2, 16, 8)

                glPopMatrix()

        glEndList()
        gluDeleteQuadric(quadric)

    def appendMolecule(self, coords, aonums):
        new_index = len(self.bond_call_lists)
        self.insertMolecule(new_index, coords, aonums)

    def insertMolecule(self, index, coords, aonums):
        self.makeCurrent()

        # create lists for atoms
        self.updateAtomCallLists(aonums)

        list_start = glGenLists(1)
        self.bond_call_lists[index] = list_start
        self.updateBondCallList(index, coords, aonums)

        self.atomcoords[index] = coords
        self.atomnos[index] = aonums

        self.updateGL()

    def updateMolecule(self, row, col, value):
        current = self.current

        assert row < len(self.atomnos[current])
        assert current < len(self.bond_call_lists)

        atomnos = self.atomnos[current]
        atomcoords = self.atomcoords[current]

        if col == 0:
            atomnos[row] = value

        elif col > 0:
            atomcoords[row][col - 1] = value
            
        self.updateAtomCallLists(atomnos)
        self.updateBondCallList(current, atomcoords, atomnos)
        self.updateGL()
        
    def centerMolecule(self, index):
        current = self.current
        atomcoords = self.atomcoords[current]
        atomnos = self.atomnos[current]

        atomcoords -= atomcoords[index, :]

        self.updateBondCallList(current, atomcoords, atomnos)
        self.updateGL()

    def saveFile(self, filename, format):
        from platform import platform
        import os
        from os.path import exists

        if "Darwin" in platform():
            print("This is a mac")

            from convertors import saveMolecule
            retval = saveMolecule(str(filename), format, self.atomnos[self.current], 
                                            self.atomcoords[self.current])

        else:
            from cclib.bridge import makeopenbabel
            import openbabel

            obmol = makeopenbabel(self.atomcoords[self.current], self.atomnos[self.current])
            obconv = openbabel.OBConversion()
            ok = obconv.SetOutFormat(format)

            if not ok:
                messageBox = QtGui.QMessageBox()
                messageBox.setIcon(QtGui.QMessageBox.Warning)
                messageBox.addButton(QtGui.QMessageBox.Ok)
                messageBox.setText("There is a problem loading the OpenBabel formats.")
                messageBox.exec_()
                return

            else:
                retval = obconv.WriteFile(obmol, str(filename))

        if not retval:
            print "Problem writing file."
            messageBox = QtGui.QMessageBox()
            messageBox.setIcon(QtGui.QMessageBox.Warning)
            messageBox.addButton(QtGui.QMessageBox.Ok)
            messageBox.setText("There is a problem saving the file.")
            messageBox.exec_()

    def initializeGL(self):

        glEnable(GL_CULL_FACE)
        
        glDepthFunc(GL_LESS)
        glEnable(GL_DEPTH_TEST)

        glEnable(GL_NORMALIZE)
        glEnable(GL_BLEND)
        glEnable(GL_LIGHTING)
        glShadeModel(GL_SMOOTH)

        #set ambient and diffuse light colors by glColor
        glEnable(GL_COLOR_MATERIAL)
        glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE) 

        #set intensity of light factors before using glColor
        specularLight = [1.0, 1.0, 1.0, 1.0]
        ambientLight = [0.2, 0.2, 0.2, 1.0]
        diffuseLight = [0.6, 0.6, 0.6, 1.0]
        position = (0.7, 0.8, 0.4, 0.0)

        #set light parameters
        glLightfv(GL_LIGHT0, GL_SPECULAR, specularLight)
        glLightfv(GL_LIGHT0, GL_AMBIENT, ambientLight)
        glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuseLight)
        glLightfv(GL_LIGHT0, GL_POSITION, position)
        glEnable(GL_LIGHT0)

        #set relfective properties
        specReflect = [ 0.8, 0.8, 0.8, 1.0]
        glMaterialfv(GL_FRONT, GL_SPECULAR, specReflect)
        glMateriali(GL_FRONT, GL_SHININESS, 45)

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glMatrixMode(GL_MODELVIEW)

# create call lists for axes
        width = 0.06
        length = 0.4

        self.axes = glGenLists(1)
        glNewList(self.axes, GL_COMPILE)

        quadric = gluNewQuadric()
        gluQuadricNormals(quadric, GLU_SMOOTH)

        glColor3f(1.0, 1.0, 1.0)
        gluSphere(quadric,width,32,32)

# draw x-axis
        glPushMatrix()
        glRotatef(90, 0.0, 1.0, 0.0)
        glColor3f(1.0, 0.0, 0.0)
        gluCylinder(quadric, width, width, length, 32, 32)
        glPushMatrix()
        glTranslate(0.0, 0.0, length)
        gluCylinder(quadric, width, 0.0, length/4, 32, 32)
        glPopMatrix()
        glPopMatrix()

# draw y-axis
        glPushMatrix()
        glRotatef(-90, 1.0, 0.0, 0.0)
        glColor3f(0.0, 1.0, 0.0)
        gluCylinder(quadric, width, width, length, 32, 32)
        glPushMatrix()
        glTranslate(0.0, 0.0, length)
        gluCylinder(quadric, width, 0.0, length/4, 32, 32)
        glPopMatrix()
        glPopMatrix()

# draw z-axis
        glPushMatrix()
        glColor3f(0.0, 0.0, 1.0)
        gluCylinder(quadric, width, width, length, 32, 32)
        glPushMatrix()
        glTranslate(0.0, 0.0, length)
        gluCylinder(quadric, width, 0.0, length/4, 32, 32)
        glPopMatrix()
        glPopMatrix()

# finish list
        gluDeleteQuadric(quadric)
        glEndList()


    def resizeGL(self, width, height):
        side = min(width, height)
        glViewport(0,0,width,height)

        glMatrixMode(GL_PROJECTION)

        if width > 0:
            glLoadIdentity()
            gluPerspective(60.0, (width+0.0)/height, 0.1, 100.0)

        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

    def renderFrame(self):
        # setup FBO to render to
        format = QtOpenGL.QGLFramebufferObjectFormat()
        format.setAttachment(QtOpenGL.QGLFramebufferObject.Depth)
        fbo = QtOpenGL.QGLFramebufferObject(self.width(), self.height(), format)

        # bind, draw, and release
        fbo.bind()
        self.draw()
        fbo.release()

        # return resulting image
        image = fbo.toImage()
        return image

    def paintGL(self):
        self.draw()

    def draw(self):

        self.makeCurrent()

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        glLoadIdentity()

# draw axes
        y = -5.0 * tan(30.0 / 180 * pi)
        glTranslatef(-1.0 * self.width() / self.height() * y, y, -6.0)

        glRotatef(self.xrot / 16.0, 1.0, 0.0, 0.0)
        glRotatef(self.yrot / 16.0, 0.0, 1.0, 0.0)
        glRotatef(self.zrot / 16.0, 0.0, 0.0, 1.0)

        glCallList(self.axes)

# now draw molecule
        current = self.current
        if current >= len(self.bond_call_lists) or current not in self.atomcoords:
            return

        glLoadIdentity()
        glPushMatrix()
        glTranslatef(self.xpos, self.ypos, self.zpos)
        glRotatef(self.xrot / 16.0, 1.0, 0.0, 0.0)
        glRotatef(self.yrot / 16.0, 0.0, 1.0, 0.0)
        glRotatef(self.zrot / 16.0, 0.0, 0.0, 1.0)

        # draw atoms
        atomcoords = self.atomcoords[current]
        atomnos = self.atomnos[current]
        atoms = zip(atomcoords, atomnos)

        for atomcoord, atomno in atoms:
            glPushMatrix()
            glTranslate(atomcoord[0], atomcoord[1], atomcoord[2])
            glCallList(self.atom_call_lists[atomno][0])
            glPopMatrix()

        # draw bonds
        bond_i = self.bond_call_lists[current]
        glCallList(bond_i)

        # draw selected atom
        for i in self.selected:
            atomcoord, atomno = atoms[i]
            glPushMatrix()
            glTranslate(atomcoord[0], atomcoord[1], atomcoord[2])
            glCallList(self.atom_call_lists[atomno][1])
            glPopMatrix()

        glPopMatrix()

    def length(self, x, y, z):
        return sqrt(x**2 + y**2 + z**2)

if __name__ == '__main__':
    import sys
    from cclib.parser import ccopen

    app = QtGui.QApplication(sys.argv)
    parser = ccopen(sys.argv[1])
    data = parser.parse()
    window = GLWidget()
    window.appendMolecule(data.atomcoords[0], data.atomnos)
    window.resize(800,600)
    window.show()
    sys.exit(app.exec_())

