
# transitions_treeview.py -- code for the Transitions View/Model of QMForge
#
# Copyright (c) Adam Tenderholt, atenderholt@gmail.com, 2014
#
# This program is free software; you can redistribute it and/or modify  
# it under the terms of the GNU General Public License as published by 
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version with the exceptions noted in the 
# Help dialog.

from PyQt4 import QtCore, QtGui
import numpy
from cclib.parser import utils

def orbitalNumberAsString(num, homo):
    if num == homo:
        text = "HOMO"
    elif num == homo + 1:
        text = "LUMO"
    elif num < homo: # occupied
        text = "H-%i" % (homo - num)
    else: # virtuals
        text = "L+%i" % (num - homo - 1)

    return text

def formatTransitionContribution(entry, homos):
    """Format an entry of a TD-DFT/CIS contribution.
       
       For example, entry=((31, 0), (36, 0), 0.206) and homo = 35 should become 
         4.2% H-4 => LUMO.

    """
    label = { 0: "(A)", 1: "(B)" }

    percent = entry[2] ** 2 * 100
    if percent >= 10.0:
        result = "%.1f%% " % percent
    elif percent >= 1.0:
        result = "%.2f%% " % percent
    else:
        return None
        
    result += "%s" % orbitalNumberAsString(entry[0][0], homos[ entry[0][1] ])
    if len(homos) == 2:
        result += label[ entry[0][1] ]
    result += " => "
    result += "%s" % orbitalNumberAsString(entry[1][0], homos[ entry[1][1] ])
    if len(homos) == 2:
        result += label[ entry[1][1] ]

    return result

def convertETInformationToDictionary(etenergies, etoscs, etsecs, homos):
    """ Take ET Info and convert it to dictionary, e.g.

        results[0] = [0, etenergies[0], etenergies[0] in nm, etoscs[0], contribution[0],
                        remaining contribs...]

    """

    results = {}
    for i in range(len(etenergies)):
        results[i] = [i]
        results[i].append(etenergies[i])
        results[i].append( utils.convertor(etenergies[i], "cm-1", "nm") )
        results[i].append(etoscs[i])
        
        # now for the hard part (sorted contributions)
        contribs = numpy.array( map(lambda x: x[2], etsecs[i]) ) ** 2
        indices = numpy.argsort(contribs)
        
        j = indices[-1] # largets value
        results[i].append( formatTransitionContribution(etsecs[i][j], homos) ) # first entry on top
        results[i].append([])
        for j in indices[-2::-1]:
            result = formatTransitionContribution(etsecs[i][j], homos)
            if result:
                results[i][-1].append(result)

    return results

class TransitionsTreeView(QtGui.QTreeWidget):
    def __init__(self, parent=None):
        super(TransitionsTreeView, self).__init__(parent)

        self.headerItem().setText(0, "#")
        self.headerItem().setText(1, "Energy (cm-1)")
        self.headerItem().setText(2, "Wavelength (nm)")
        self.headerItem().setText(3, "Osc. Strength")
        self.headerItem().setText(4, "Contributions")

        self.results = None

    def setResults(self, results):
        self.results = results
        for i in range(len(results)):
            transition = QtGui.QTreeWidgetItem(self)
            transition.setText(0, "%i" % (results[i][0] + 1))
            transition.setText(1, "%i" % results[i][1])
            transition.setText(2, "%.1f" % results[i][2])
            transition.setText(3, "%.4f" % results[i][3])
            transition.setText(4, results[i][4] + " ...")
            for j in range(len(results[i][-1])):
                contrib = QtGui.QTreeWidgetItem(transition)
                contrib.setText(4, results[i][-1][j])

if __name__ == "__main__":
    import sys
    from cclib.parser import ccopen

    app = QtGui.QApplication(sys.argv)

    parser = ccopen(sys.argv[1])
    data = parser.parse()

    results = convertETInformationToDictionary(data.etenergies, data.etoscs, data.etsecs, 
                                                data.homos)

    view = TransitionsTreeView()
    view.setResults(results)
    view.show()

    sys.exit(app.exec_())    

