
# transitions_widget.py -- code for the Transitions Widget of QMForge
#
# Copyright (c) Adam Tenderholt, atenderholt@gmail.com, 2014
#
# This program is free software; you can redistribute it and/or modify  
# it under the terms of the GNU General Public License as published by 
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version with the exceptions noted in the 
# Help dialog.

import os

from PyQt4 import QtCore, QtGui, Qwt5
import numpy
from cclib.parser import utils

from ui_transitions_widget import Ui_transitionWidget
from transitions_treeview import convertETInformationToDictionary

labels = { "nm": "Wavelength (nm)", "cm-1": "Energy (cm-1)", "eV": "Energy (ev)" }

class TransitionsWidget(QtGui.QWidget, Ui_transitionWidget):

    def __init__(self, parent=None):
        super(TransitionsWidget, self).__init__(parent)

        self.setupUi(self)
        self.data = None
        self.units = str(self.comboBox.currentText())

        xlabel = labels[self.units]
        self.qwtPlot.setAxisTitle(Qwt5.QwtPlot.xBottom, xlabel)
        self.qwtPlot.setAxisTitle(Qwt5.QwtPlot.yLeft, "Molar Absorptivity")
        self.qwtPlot.setAxisTitle(Qwt5.QwtPlot.yRight, "Oscillator Strength")
        self.qwtPlot.enableAxis(Qwt5.QwtPlot.yRight)

        self.setAcceptDrops(True)

    def dragEnterEvent(self, event):
        event.acceptProposedAction()

    def dragMoveEvent(self, event):
        event.acceptProposedAction()

    def dropEvent(self, event):
        mimedata = event.mimeData()
        for url in mimedata.urls():
            if url.isLocalFile():
                filename = str(url.toLocalFile())
                custom_data = numpy.loadtxt(filename)
                self.exp_curve = Qwt5.QwtPlotCurve()
                self.exp_curve.attach(self.qwtPlot)
                self.exp_curve.setData(custom_data[:, 0], custom_data[:, 1])
                self.exp_curve.setAxis(Qwt5.QwtPlot.xBottom, Qwt5.QwtPlot.yRight)
                self.qwtPlot.setAxisScale(Qwt5.QwtPlot.yRight, 0, max(custom_data[:, 1]))
                self.qwtPlot.replot()

        event.acceptProposedAction()

    def setData(self, data):
        self.data = data
        self.resetDefaults()
        self.results = convertETInformationToDictionary(data.etenergies, data.etoscs, data.etsecs, 
                                                        data.homos)
        self.treeWidget.setResults(self.results)

    def resetDefaults(self):
        if self.data and hasattr(self.data, "etenergies"):
            units = str(self.comboBox.currentText())
            fwhm = 3000
            temp_min = min(self.data.etenergies) - 3 * fwhm
            temp_max = max(self.data.etenergies) + 3 * fwhm

            if units == "nm":
                xmin = utils.convertor(temp_max, "cm-1", "nm")
                xmax = utils.convertor(temp_min, "cm-1", "nm")
                if xmin > 1 or xmax > 1:
                    label = "%i"
                else:
                    label = "%.6f"
            elif units == "eV":
                xmin = utils.convertor(temp_min, "cm-1", "eV")
                xmax = utils.convertor(temp_max, "cm-1", "eV")
                label = "%.2f"
            else:
                xmin = temp_min
                xmax = temp_max
                label = "%i"

            self.startLineEdit.setText(label % xmin)
            self.endLineEdit.setText(label % xmax)
            self.FWHMLineEdit.setText("%i" % fwhm)
            self.offsetLineEdit.setText("0")
            self.updatePlot()

    def updatePlotForUnits(self, text):
        units = str(text)
        xlabel = labels[units]
        self.qwtPlot.setAxisTitle(Qwt5.QwtPlot.xBottom, xlabel)

        temp_min = float(self.startLineEdit.text())
        temp_max = float(self.endLineEdit.text())

        if self.units == "nm" or units == "nm":
            xmax = utils.convertor(temp_min, self.units, units)
            xmin = utils.convertor(temp_max, self.units, units)
        else:
            xmin = utils.convertor(temp_min, self.units, units)
            xmax = utils.convertor(temp_max, self.units, units)

        if units == "eV":
            self.startLineEdit.setText("%.2f" % xmin)
            self.endLineEdit.setText("%.2f" % xmax)
        else:
            self.startLineEdit.setText("%i" % xmin)
            self.endLineEdit.setText("%i" % xmax)

        self.units = units

        self.updatePlot()

    def updatePlot(self):
        if not self.data:
            return

        units_min = float(self.startLineEdit.text())
        units_max = float(self.endLineEdit.text())
        fwhm = float(self.FWHMLineEdit.text())
        units_offset = float(self.offsetLineEdit.text())

        self.qwtPlot.setAxisScale(Qwt5.QwtPlot.xBottom, units_min, units_max)

        # generate xvalues, first converting to cm-1 if necessary
        if str(self.comboBox.currentText()) == "nm":
            print units_min, units_max
            xmin = utils.convertor(units_max, "nm", "cm-1")
            xmax = utils.convertor(units_min, "nm", "cm-1")
        elif str(self.comboBox.currentText()) == "eV":
            xmin = utils.convertor(units_min, "eV", "cm-1")
            xmax = utils.convertor(units_max, "eV", "cm-1")
            offset = utils.convertor(units_offset, "eV", "cm-1")
        else:
            xmin = units_min
            xmax = units_max
            offset = units_offset

        # use a sensible default, based on width of peaks
        xvalues = numpy.arange(xmin, xmax, fwhm / 10.0)

        energies = self.data.etenergies     # this should be in cm-1
        oscs = self.data.etoscs 

        # calculate spectrum!
        self.yvalues = numpy.zeros(xvalues.shape)
        for energy, height in zip(energies, oscs * 2.174e8 / fwhm):
            if str(self.comboBox.currentText()) == "nm": # need to calculate on the fly
                temp_energy = utils.convertor(energy, "cm-1", "nm")
                offset = utils.convertor(temp_energy + units_offset, "nm", "cm-1") - energy

            self.yvalues += self.calcSpectrum(xvalues, energy + offset, fwhm, height)

        if not hasattr(self, "curve"):
            self.curve = Qwt5.QwtPlotCurve()
            self.curve.attach(self.qwtPlot)
            self.sticks = Qwt5.QwtPlotCurve()
            self.sticks.setStyle(Qwt5.QwtPlotCurve.Sticks)
            self.sticks.setAxis(Qwt5.QwtPlot.xBottom, Qwt5.QwtPlot.yRight)
            self.sticks.attach(self.qwtPlot)

        # convert xvalues in cm-1 to units
        if str(self.comboBox.currentText()) == "nm":
            self.xvalues = numpy.array( map(lambda x: utils.convertor(x, "cm-1", "nm"), xvalues) )
            self.xpoints = numpy.array( map(lambda x: utils.convertor(x, "cm-1", "nm"), 
                                        energies) )
        elif str(self.comboBox.currentText()) == "eV":
            self.xvalues = numpy.array( map(lambda x: utils.convertor(x, "cm-1", "eV"), xvalues) )
            self.xpoints = numpy.array( map(lambda x: utils.convertor(x, "cm-1", "eV"), 
                                        energies) )
        else:
            self.xvalues = xvalues
            self.xpoints = energies

        self.xpoints += units_offset
        self.curve.setData(self.xvalues, self.yvalues)
        self.sticks.setData(self.xpoints, oscs)

        self.qwtPlot.replot()

    def calcSpectrum(self, xvalues, center, width, height):
        A = -2.7726 / width ** 2
        return height * numpy.exp(A * numpy.power(xvalues - center, 2))

    def changeCurrentItem(self, current, previous):
        try:
            index = int(current.text(0)) - 1
        except ValueError:
            index = int(current.parent().text(0)) - 1

        if not hasattr(self, "selectedCurve"):
            self.selectedCurve = Qwt5.QwtPlotCurve()
            self.selectedCurve.setStyle(Qwt5.QwtPlotCurve.Sticks)
            self.selectedCurve.setAxis(Qwt5.QwtPlot.xBottom, Qwt5.QwtPlot.yRight)
            pen = QtGui.QPen(QtGui.QColor("red"))
            pen.setWidth(2)
            self.selectedCurve.setPen(pen)
            self.selectedCurve.attach(self.qwtPlot)

        xvalues = [self.xpoints[index]]
        self.selectedCurve.setData(xvalues, [self.data.etoscs[index]])
        self.qwtPlot.replot()

    def saveSpectrum(self, origfilename):
        if not hasattr(self, "data"):
            print "Data hasn't been set"

        if not origfilename:
            QtGui.QMessageBox.critical(self, "Error!", "Cannot save without original filename!")
            return

        suggest = os.path.splitext(origfilename)[0] + "_spectrum.txt"
        filename = QtGui.QFileDialog.getSaveFileName(self, "Save excited state spectrum...", \
                                suggest, "*.txt")

        oscs = self.data.etoscs
        file = open(filename, "w")
        file.write("Energy\tAbsorptivity\tEnergy\tOsc.Strength\tFWHM=%i\tOffset=%.4f\n" % \
                    (int(self.FWHMLineEdit.text()), float(self.offsetLineEdit.text())))
        small_size = len(self.xpoints)
        for i in range(small_size):
            file.write("%8.3f\t%6.3e\t%8.3f\t%6.3e\n" % (self.xvalues[i], self.yvalues[i],
                                                        self.xpoints[i], oscs[i]))
        combined = numpy.zeros((len(self.xvalues), 2))
        combined[:, 0] = self.xvalues
        combined[:, 1] = self.yvalues
        numpy.savetxt(file, combined[small_size:], fmt="%8.3f\t%6.3e")

        file.close()

    def saveData(self, origfilename):
        if not hasattr(self, "results"):
            mesg = "There is a problem saving data from the Transitions tab."
            QtGui.QMessageBox.critical(self, "Error!", mesg)
            return

        if not origfilename:
            QtGui.QMessageBox.critical(self, "Error!", "Cannot save without original filename!")
            return

        suggest = os.path.splitext(origfilename)[0] + "_contributions.txt"
        filename = QtGui.QFileDialog.getSaveFileName(self, "Save excited state contributions...", \
                                suggest, "*.txt")


        file = open(filename, "w")
        file.write("Index\tEnergy.cm^-1\tWavelength\tOsc.Strength\tContributions\n")
        for i in range(len(self.results)): # dictionary, with keys of integers
            entry = self.results[i]
            file.write("%-2i\t%8.3f\t%8.3f\t%6.3e\t%s" % (entry[0], entry[1], entry[2], entry[3],
                                                            entry[4]))
            for j in range(len(entry[5])):
                file.write(", %s" % (entry[5][j]))

            file.write("\n")

        file.close()


if __name__ == "__main__":
    import sys
    from cclib.parser import ccopen

    parser = ccopen(sys.argv[1])
    data = parser.parse()

    app = QtGui.QApplication(sys.argv)

    widget = TransitionsWidget()
    widget.setData(data)
    widget.show()

    result = app.exec_()
    widget.saveSpectrum(sys.argv[1])
    widget.saveData(sys.argv[1])

    sys.exit(result)

